"""This module defines util classes
which allow to execute operations
for each element of a list of objects of the same class.
"""

import abc
import collections.abc
import functools


class SubclassingMeta(abc.ABCMeta):
    """Metaclass which creates a subclass for each instance.

    As decorators only work
    if the class object contains the declarations,
    we need to create a subclass for each different type
    if we want to dynamically use them.
    """

    SUBCLASS_IDENTIFIER = "__subclassed__"

    def __call__(cls, *args, **kwargs):
        if hasattr(cls, SubclassingMeta.SUBCLASS_IDENTIFIER):
            return super().__call__(*args, **kwargs)

        subclass = type(
            cls.__name__,
            (cls,),
            {
                SubclassingMeta.SUBCLASS_IDENTIFIER: SubclassingMeta.SUBCLASS_IDENTIFIER
            },
        )
        return subclass(*args, **kwargs)


class BatchList(collections.abc.MutableSequence, metaclass=SubclassingMeta):
    """BatchList provides the execution of methods and field access
    for each element of a list of instances of the same class
    in a similar way to one of these instances it would.
    """

    __attributes_declared = False

    class BatchMember:
        def __init__(self, outer, name):
            """
            Args:
                outer (BatchList): Outer class instance
            """
            self.outer = outer
            self.name = name

    class BatchField(BatchMember):
        def __get__(self, owner_instance, owner_class):
            return BatchList(
                [
                    instance.__getattribute__(self.name)
                    for instance in self.outer
                ]
            )

        def __set__(self, owner_instance, value):
            for instance in self.outer:
                instance.__setattr__(self.name, value)

        def __delete__(self, instance):
            for instance in self.outer:
                instance.__delattr__(self.name)

    class BatchMethod(BatchMember):
        def __call__(self, *args, **kwargs):
            return BatchList(
                [
                    instance.__getattribute__(self.name)(*args, **kwargs)
                    for instance in self.outer
                ]
            )

    def __init__(self, collection: list):
        """
        Args:
            collection (List): List of target instances
        """
        self.__collection = collection.copy()
        self.__initialized = False
        self.__type = None
        self.entered = False
        self.__attributes_declared = True
        self.__init_members__()

    def __call__(self, *args, **kwargs):
        if self.__initialized:
            raise TypeError("'%s' object is not callable" % self.__type)
        return BatchList([])

    def __getattr__(self, name):
        if self.__initialized:
            return AttributeError(
                "'%s' object has no attribute '%s'" % (self.__type, name)
            )
        return BatchList([])

    def __setattr__(self, name, value):
        if (
            not self.__attributes_declared
            or self.__initialized
            or not isinstance(getattr(self, name), BatchList)
        ):
            super().__setattr__(name, value)

    def __init_members__(self):
        if self.__collection and not self.__initialized:
            # Note: We can't simply use the class,
            #       as the attributes exists only after the instantiation
            self.__initialized = True
            instance = self.__collection[0]
            self.__type = type(instance)
            self.__init_attributes__(instance)
            self.__init_methods__(instance)

    def __declare_decorator__(self, name, decorator):
        setattr(type(self), name, decorator)

    def __init_attributes__(self, target_instance):
        for name in self.__get_public_attributes(target_instance):
            self.__declare_decorator__(name, BatchList.BatchField(self, name))

    @staticmethod
    def __get_public_attributes(target_instance):
        attributes = (
            vars(target_instance)
            if hasattr(target_instance, "__dict__")
            else []
        )
        return (name for name in attributes if not name.startswith("_"))

    @staticmethod
    @functools.lru_cache()
    def __get_public_members(target_type):
        members = {
            name: member
            for type_members in map(vars, reversed(target_type.mro()))
            for name, member in type_members.items()
        }
        return {
            name: member
            for name, member in members.items()
            if not name.startswith("_")
        }

    def __init_methods__(self, target_instance):
        public_members = self.__get_public_members(type(target_instance))
        for name, value in public_members.items():
            if callable(value):
                self.__declare_decorator__(
                    name, BatchList.BatchMethod(self, name)
                )
            else:
                # should be an decorator
                self.__declare_decorator__(
                    name, BatchList.BatchField(self, name)
                )

    def __enter__(self):
        self.entered = True
        return BatchList([instance.__enter__() for instance in self])

    def __exit__(self, *args):
        for instance in self:
            instance.__exit__(*args)

    def __iadd__(self, other):
        if self.entered:
            for i in other:
                i.__enter__()
        self.__collection.__iadd__(other)
        self.__init_members__()
        return self

    def append(self, item):
        if self.entered:
            item.__enter__()
        self.__collection.append(item)
        self.__init_members__()

    def insert(self, index, item):
        if self.entered:
            item.__enter__()
        self.__collection.insert(index, item)
        self.__init_members__()

    def extend(self, iterable):
        for item in iterable:
            self.append(item)

    def __add__(self, other):
        return BatchList(self.__collection.__add__(other))

    def reverse(self):
        self.__collection.reverse()

    def clear(self):
        if self.entered:
            for i in self.__collection:
                i.__exit__(None, None, None)
        self.__collection.clear()

    def copy(self):
        return BatchList(self.__collection.copy())

    def pop(self, *args):
        result = self.__collection.pop(*args)

        if self.entered:
            result.__exit__(None, None, None)

        return result

    def remove(self, value):
        if self.entered:
            value.__exit__(None, None, None)
        return self.__collection.remove(value)

    def __isub__(self, other):
        for i in other:
            self.remove(i)
        return self

    def __sub__(self, other):
        copied = self.copy()
        copied -= other
        return copied

    def __len__(self):
        return len(self.__collection)

    def __delitem__(self, key):
        return self.pop(key)

    def __setitem__(self, key, value):
        self.pop(key)
        self.insert(key, value)

    def __getitem__(self, key):
        return self.__collection[key]

    def count(self, *args, **kwargs):
        return self.__collection.count(*args, **kwargs)

    def index(self, *args, **kwargs):
        return self.__collection.index(*args, **kwargs)

    def __iter__(self):
        return iter(self.__collection)

    def __contains__(self, item):
        return item in self.__collection

    def __reversed__(self):
        return reversed(self.__collection)


if __name__ == "__main__":

    class FooBar:
        def __init__(self, a, b, c):
            self.mhm = a
            self.b = b
            self.c = c

        def ok(self):
            return self.b

        @property
        def prop(self):
            return self.c

    # print attributes
    # print(vars(FooBar()))
    # print properties and methods
    # print(vars(FooBar).keys())
    blist = BatchList([FooBar("foo", "bar", "yay")])
    blist += [FooBar("foobar", "barfoo", "yay foobar")]
    print("mhm", blist.mhm)
    print("prop", blist.prop)
    # print('ok', blist.ok)
    print("ok call", blist.ok())



"""This module contains user interface related classes and methods.
"""

import abc
import weakref
import attr

import PIL.Image as Image

import ueberzug.xutil as xutil
import ueberzug.geometry as geometry
import ueberzug.scaling as scaling
import ueberzug.X as X


def roundup(value, unit):
    return ((value + (unit - 1)) & ~(unit - 1)) >> 3


class WindowFactory:
    """Window factory class"""

    def __init__(self, display):
        self.display = display

    @abc.abstractmethod
    def create(self, *window_infos: xutil.TerminalWindowInfo):
        """Creates a child window for each window id."""
        raise NotImplementedError()


class CanvasWindow(X.OverlayWindow):
    """Ensures unmapping of windows"""

    class Factory(WindowFactory):
        """CanvasWindows factory class"""

        def __init__(self, display, view):
            super().__init__(display)
            self.view = view

        def create(self, *window_infos: xutil.TerminalWindowInfo):
            return [
                CanvasWindow(self.display, self.view, info)
                for info in window_infos
            ]

    class Placement:
        @attr.s
        class TransformedImage:
            """Data class which contains the options
            an image was transformed with
            and the image data."""

            options = attr.ib(type=tuple)
            data = attr.ib(type=bytes)

        def __init__(
            self,
            x: int,
            y: int,
            width: int,
            height: int,
            scaling_position: geometry.Point,
            scaler: scaling.ImageScaler,
            path: str,
            image: Image,
            last_modified: int,
            cache: weakref.WeakKeyDictionary = None,
        ):
            # x, y are useful names in this case
            # pylint: disable=invalid-name
            self.x = x
            self.y = y
            self.width = width
            self.height = height
            self.scaling_position = scaling_position
            self.scaler = scaler
            self.path = path
            self.image = image
            self.last_modified = last_modified
            self.cache = cache or weakref.WeakKeyDictionary()

        def transform_image(
            self,
            term_info: xutil.TerminalWindowInfo,
            width: int,
            height: int,
            format_scanline: tuple,
        ):
            """Scales to image and calculates
            the width & height needed to display it.

            Returns:
                tuple of (width: int, height: int, image: bytes)
            """
            image = self.image.await_image()
            scanline_pad, scanline_unit = format_scanline
            transformed_image = self.cache.get(term_info)
            final_size = self.scaler.calculate_resolution(image, width, height)
            options = (
                self.scaler.get_scaler_name(),
                self.scaling_position,
                final_size,
            )

            if (
                transformed_image is None
                or transformed_image.options != options
            ):
                image = self.scaler.scale(
                    image, self.scaling_position, width, height
                )
                stride = roundup(image.width * scanline_unit, scanline_pad)
                transformed_image = self.TransformedImage(
                    options, image.tobytes("raw", "BGRX", stride, 0)
                )
                self.cache[term_info] = transformed_image

            return (*final_size, transformed_image.data)

        def resolve(
            self,
            pane_offset: geometry.Distance,
            term_info: xutil.TerminalWindowInfo,
            format_scanline,
        ):
            """Resolves the position and size of the image
            according to the teminal window information.

            Returns:
                tuple of (x: int, y: int, width: int, height: int,
                          image: PIL.Image)
            """
            # x, y are useful names in this case
            # pylint: disable=invalid-name
            image = self.image.await_image()
            x = int(
                (self.x + pane_offset.left) * term_info.font_width
                + term_info.padding_horizontal
            )
            y = int(
                (self.y + pane_offset.top) * term_info.font_height
                + term_info.padding_vertical
            )
            width = int(
                (self.width and (self.width * term_info.font_width))
                or image.width
            )
            height = int(
                (self.height and (self.height * term_info.font_height))
                or image.height
            )

            return (
                x,
                y,
                *self.transform_image(
                    term_info, width, height, format_scanline
                ),
            )

    def __init__(
        self, display: X.Display, view, parent_info: xutil.TerminalWindowInfo
    ):
        """Changes the foreground color of the gc object.

        Args:
            display (X.Display): any created instance
            parent_id (int): the X11 window id of the parent window
        """
        super().__init__(display, parent_info.window_id)
        self.parent_info = parent_info
        self._view = view
        self.scanline_pad = display.bitmap_format_scanline_pad
        self.scanline_unit = display.bitmap_format_scanline_unit
        self.screen_width = display.screen_width
        self.screen_height = display.screen_height
        self._image = X.Image(display, self.screen_width, self.screen_height)

    def __enter__(self):
        self.draw()
        return self

    def __exit__(self, *args):
        pass

    def draw(self):
        """Draws the window and updates the visibility mask."""
        rectangles = []

        if not self.parent_info.ready:
            self.parent_info.calculate_sizes(self.width, self.height)

        for placement in self._view.media.values():
            # x, y are useful names in this case
            # pylint: disable=invalid-name
            x, y, width, height, image = placement.resolve(
                self._view.offset,
                self.parent_info,
                (self.scanline_pad, self.scanline_unit),
            )
            rectangles.append((x, y, width, height))
            self._image.draw(x, y, width, height, image)

        self._image.copy_to(
            self.id,
            0,
            0,
            min(self.width, self.screen_width),
            min(self.height, self.screen_height),
        )
        self.set_visibility_mask(rectangles)
        super().draw()

    def reset_terminal_info(self):
        """Resets the terminal information of this window."""
        self.parent_info.reset()



import select
import fcntl
import contextlib
import pathlib


class LineReader:
    """Async iterator class used to read lines"""

    def __init__(self, loop, file):
        self._loop = loop
        self._file = file

    @staticmethod
    async def read_line(loop, file):
        """Waits asynchronously for a line and returns it"""
        return await loop.run_in_executor(None, file.readline)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if select.select([self._file], [], [], 0)[0]:
            return self._file.readline()
        return await LineReader.read_line(self._loop, self._file)


@contextlib.contextmanager
def lock(path: pathlib.PosixPath):
    """Creates a lock file,
    a file protected from beeing used by other processes.
    (The lock file isn't the same as the file of the passed path.)

    Args:
        path (pathlib.PosixPath): path to the file
    """
    path = path.with_suffix(".lock")

    if not path.exists():
        path.touch()

    with path.open("r+") as lock_file:
        try:
            fcntl.lockf(lock_file.fileno(), fcntl.LOCK_EX)
            yield lock_file
        finally:
            fcntl.lockf(lock_file.fileno(), fcntl.LOCK_UN)



import subprocess
import shlex
import os

import ueberzug.geometry as geometry


def is_used():
    """Determines whether this program runs in tmux or not."""
    return get_pane() is not None


def get_pane():
    """Determines the pane identifier this process runs in.

    Returns:
        str or None
    """
    return os.environ.get("TMUX_PANE")


def get_session_id():
    """Determines the session identifier this process runs in.

    Returns:
        str
    """
    return (
        subprocess.check_output(
            ["tmux", "display", "-p", "-F", "#{session_id}", "-t", get_pane()]
        )
        .decode()
        .strip()
    )


def get_offset():
    """Determines the offset
    of the pane (this process runs in)
    within it's tmux window.
    """
    result = subprocess.check_output(
        [
            "tmux",
            "display",
            "-p",
            "-F",
            "#{pane_top},#{pane_left},"
            "#{pane_bottom},#{pane_right},"
            "#{window_height},#{window_width}",
            "-t",
            get_pane(),
        ]
    ).decode()
    top, left, bottom, right, height, width = (
        int(i) for i in result.split(",")
    )
    return geometry.Distance(top, left, height - bottom, width - right)


def is_window_focused():
    """Determines whether the window
    which owns the pane
    which owns this process is focused.
    """
    result = subprocess.check_output(
        [
            "tmux",
            "display",
            "-p",
            "-F",
            "#{window_active},#{pane_in_mode}",
            "-t",
            get_pane(),
        ]
    ).decode()
    return result == "1,0\n"


def get_client_pids():
    """Determines the tty for each tmux client
    displaying the pane this program runs in.
    """
    if not is_window_focused():
        return {}

    return {
        int(pid)
        for pid in subprocess.check_output(
            ["tmux", "list-clients", "-F", "#{client_pid}", "-t", get_pane()]
        )
        .decode()
        .splitlines()
    }


def register_hook(event, command):
    """Updates the hook of the passed event
    for the pane this program runs in
    to the execution of a program.

    Note: tmux does not support multiple hooks for the same target.
    So if there's already an hook registered it will be overwritten.
    """
    subprocess.check_call(
        [
            "tmux",
            "set-hook",
            "-t",
            get_pane(),
            event,
            "run-shell " + shlex.quote(command),
        ]
    )


def unregister_hook(event):
    """Removes the hook of the passed event
    for the pane this program runs in.
    """
    subprocess.check_call(["tmux", "set-hook", "-u", "-t", get_pane(), event])



"""This module contains x11 utils"""

import functools
import asyncio
import os

import ueberzug.tmux_util as tmux_util
import ueberzug.terminal as terminal
import ueberzug.process as process
import ueberzug.X as X


PREPARED_DISPLAYS = []
DISPLAY_SUPPLIES = 1


class Events:
    """Async iterator class for x11 events"""

    def __init__(self, loop, display: X.Display):
        self._loop = loop
        self._display = display

    @staticmethod
    async def wait_for_event(loop, display: X.Display):
        """Waits asynchronously for an x11 event and returns it"""
        return await loop.run_in_executor(None, display.wait_for_event)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await Events.wait_for_event(self._loop, self._display)


class TerminalWindowInfo(terminal.TerminalInfo):
    def __init__(self, window_id, fd_pty=None):
        super().__init__(fd_pty)
        self.window_id = window_id


@functools.lru_cache()
def get_parent_pids(pid):
    """Determines all parent pids of this process.
    The list is sorted from youngest parent to oldest parent.
    """
    pids = []
    next_pid = pid

    while next_pid > 1:
        pids.append(next_pid)
        next_pid = process.get_parent_pid(next_pid)

    return pids


def get_pid_window_id_map(display: X.Display):
    """Determines the pid of each mapped window.

    Returns:
        dict of {pid: window_id}
    """
    return {
        display.get_window_pid(window_id): window_id
        for window_id in display.get_child_window_ids()
    }


def sort_by_key_list(mapping: dict, key_list: list):
    """Sorts the items of the mapping
    by the index of the keys in the key list.

    Args:
        mapping (dict): the mapping to be sorted
        key_list (list): the list which specifies the order

    Returns:
        list: which contains the sorted items as tuples
    """
    key_map = {key: index for index, key in enumerate(key_list)}
    return sorted(
        mapping.items(), key=lambda item: key_map.get(item[0], float("inf"))
    )


def key_intersection(mapping: dict, key_list: list):
    """Creates a new map which only contains the intersection
    of the keys.

    Args:
        mapping (dict): the mapping to be filtered
        key_list (list): the keys to be used as a whitelist

    Returns:
        dict: which only contains keys which are also in key_list
    """
    key_map = {key: index for index, key in enumerate(key_list)}
    return {key: value for key, value in mapping.items() if key in key_map}


def get_first_pty(pids: list):
    """Determines the pseudo terminal of
    the first process in the passed list which owns one.
    """
    for pid in pids:
        pty_slave_file = process.get_pty_slave(pid)
        if pty_slave_file:
            return pty_slave_file

    return None


def get_parent_window_infos(display: X.Display):
    """Determines the window id of each
    terminal which displays the program using
    this layer.

    Returns:
        list of TerminalWindowInfo
    """
    window_infos = []
    client_pids = {}

    if tmux_util.is_used():
        client_pids = tmux_util.get_client_pids()
    else:
        client_pids = {process.get_own_pid()}

    if client_pids:
        pid_window_id_map = get_pid_window_id_map(display)
        # Insert current window's PID & WID to the end of map to support tabbed.
        # NOTE: Terminal (current window) must have WINDOWID as env. variable.
        if os.environ.get("WINDOWID") != None:
            pid_window_id_map[os.getpid()] = int(os.environ.get("WINDOWID"))

        for pid in client_pids:
            ppids = get_parent_pids(pid)
            ppid_window_id_map = key_intersection(pid_window_id_map, ppids)
            try:
                window_pid, window_id = next(
                    iter(sort_by_key_list(ppid_window_id_map, ppids))
                )
                # window_children_pids = ppids[:ppids.index(window_pid)][::-1]
                pty = get_first_pty(ppids)
                window_infos.append(TerminalWindowInfo(window_id, pty))
            except StopIteration:
                # Window needs to be mapped,
                # otherwise it's not listed in _NET_CLIENT_LIST
                pass

    return window_infos



"""This module defines data parser
which can be used to exchange information between processes.
"""

import json
import abc
import shlex
import itertools
import enum


class Parser:
    """Subclasses of this abstract class define
    how to input will be parsed.
    """

    @staticmethod
    @abc.abstractmethod
    def get_name():
        """Returns the constant name which is associated to this parser."""
        raise NotImplementedError()

    @abc.abstractmethod
    def parse(self, line):
        """Parses a line.

        Args:
            line (str): read line

        Returns:
            dict containing the key value pairs of the line
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def unparse(self, data):
        """Composes the data to a string
        whichs follows the syntax of the parser.

        Args:
            data (dict): data as key value pairs

        Returns:
            string
        """
        raise NotImplementedError()


class JsonParser(Parser):
    """Parses json input"""

    @staticmethod
    def get_name():
        return "json"

    def parse(self, line):
        try:
            data = json.loads(line)
            if not isinstance(data, dict):
                raise ValueError(
                    "Expected to parse an json object, got " + line
                )
            return data
        except json.JSONDecodeError as error:
            raise ValueError(error)

    def unparse(self, data):
        return json.dumps(data)


class SimpleParser(Parser):
    """Parses key value pairs separated by a tab.
    Does not support escaping spaces.
    """

    SEPARATOR = "\t"

    @staticmethod
    def get_name():
        return "simple"

    def parse(self, line):
        components = line.split(SimpleParser.SEPARATOR)

        if len(components) % 2 != 0:
            raise ValueError(
                "Expected key value pairs, "
                + "but at least one key has no value: "
                + line
            )

        return {
            key: value
            for key, value in itertools.zip_longest(
                components[::2], components[1::2]
            )
        }

    def unparse(self, data):
        return SimpleParser.SEPARATOR.join(
            str(key) + SimpleParser.SEPARATOR + str(value.replace("\n", ""))
            for key, value in data.items()
        )


class BashParser(Parser):
    """Parses input generated
    by dumping associative arrays with `declare -p`.
    """

    @staticmethod
    def get_name():
        return "bash"

    def parse(self, line):
        # remove 'typeset -A varname=( ' and ')'
        start = line.find("(")
        end = line.rfind(")")

        if not 0 <= start < end:
            raise ValueError(
                "Expected input to be formatted like "
                "the output of bashs `declare -p` function. "
                "Got: " + line
            )

        components = itertools.dropwhile(
            lambda text: not text or text[0] != "[",
            shlex.split(line[start + 1 : end]),
        )
        return {
            key[1:-1]: value
            for pair in components
            for key, value in (pair.split("=", maxsplit=1),)
        }

    def unparse(self, data):
        return " ".join(
            "[" + str(key) + "]=" + shlex.quote(value)
            for key, value in data.items()
        )


@enum.unique
class ParserOption(str, enum.Enum):
    JSON = JsonParser
    SIMPLE = SimpleParser
    BASH = BashParser

    def __new__(cls, parser_class):
        inst = str.__new__(cls)
        inst._value_ = parser_class.get_name()
        inst.parser_class = parser_class
        return inst
